import numpy, copy
from numpy import nan

from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.widgets.gui import ConfirmDialog
from oasys.util.oasys_util import TriggerIn, TriggerOut
import oasys.util.oasys_util as OU

from srxraylib.util.data_structures import ScaledMatrix

from scipy.interpolate import RectBivariateSpline

from wofrysrw.propagator.wavefront2D.srw_wavefront import SRWWavefront, PolarizationComponent, Polarization
from wofrysrw.beamline.optical_elements.other.srw_crl import SRWCRL
from wofry.propagator.propagator import PropagationManager
from wofrysrw.propagator.propagators2D.srw_fresnel_native import SRW_APPLICATION
from wofrysrw.propagator.propagators2D.srw_propagation_mode import SRWPropagationMode


from orangecontrib.srw.util.srw_objects import SRWData
from orangecontrib.srw.util.srw_util import SRWPlot
from orangecontrib.srw.widgets.gui.ow_srw_wavefront_viewer import SRWWavefrontViewer


class OWThicknessErrorPhaseShift(SRWWavefrontViewer):
    name = "Thickness Error Phase Shift"
    description = "Thickness Error Phase Shift"
    icon = "icons/thickness_phase_shifter.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 5
    category = "Display Data Tools"
    keywords = ["data", "file", "load", "read"]

    outputs = [{"name":"SRWData",
                "type":SRWData,
                "doc":"SRW Optical Element Data",
                "id":"data"},
               {"name":"Trigger",
                "type": TriggerIn,
                "doc":"Feedback signal to start a new beam simulation",
                "id":"Trigger"}]

    inputs = [("SRWData", SRWData, "set_input"),
              ("Error Profiles", list, "setErrorProfiles"),
              ("Trigger", TriggerOut, "propagate_new_wavefront")]

    crl_error_profiles = Setting([])
    crl_scaling_factor = Setting(1.0)

    TABS_AREA_HEIGHT = 555
    CONTROL_AREA_WIDTH = 405

    def __init__(self):
        super().__init__()

        self.runaction = widget.OWAction("Propagate Wavefront", self)
        self.runaction.triggered.connect(self.propagate_wavefront)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Propagate Wavefront", callback=self.propagate_wavefront)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        gui.separator(self.controlArea)

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_thick = oasysgui.createTabPage(self.tabs_setting, "Thickness Error")

        input_box = oasysgui.widgetBox(tab_thick, "Thickness Error Files", addSpace=True, orientation="vertical", height=390, width=self.CONTROL_AREA_WIDTH-20)

        self.files_area = oasysgui.textArea(height=315)

        input_box.layout().addWidget(self.files_area)

        self.refresh_files_text_area()

        oasysgui.lineEdit(input_box, self, "crl_scaling_factor", "Thickness Error Scaling Factor", labelWidth=260, valueType=float, orientation="horizontal")

    def refresh_files_text_area(self):
        text = ""

        for file in self.crl_error_profiles:
            text += file + "\n"

        self.files_area.setText(text)

    def setErrorProfiles(self, error_profiles):
        try:
            if not error_profiles is None:
                self.crl_error_profiles = error_profiles

                self.refresh_files_text_area()
        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 exception.args[0],
                                 QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def set_input(self, srw_data):
        if not srw_data is None:
            self.input_srw_data = srw_data

            if self.is_automatic_run:
                self.propagate_wavefront()

    def set_srw_live_propagation_mode(self):
        if PropagationManager.Instance().get_propagation_mode(SRW_APPLICATION)==SRWPropagationMode.WHOLE_BEAMLINE:
            raise ValueError("Propagation Mode not supported, switch to Element by Element")
        else:
            super(OWThicknessErrorPhaseShift, self).set_srw_live_propagation_mode()

    def propagate_wavefront(self):
        try:
            self.progressBarInit()

            if self.input_srw_data is None: raise Exception("No Input Data")

            self.check_data()

            input_wavefront      = self.input_srw_data.get_srw_wavefront().duplicate()
            srw_beamline         = self.input_srw_data.get_srw_beamline().duplicate()
            optical_element      = srw_beamline.get_beamline_element_at(-1).get_optical_element()
            coordinates          = srw_beamline.get_beamline_element_at(-1).get_coordinates()

            if not isinstance(optical_element, SRWCRL):
                raise ValueError("Thickness Error Phase Shift should be connected to a CRL optical element")

            if coordinates.q() != 0.0:
                raise ValueError("Thickness Error Phase Shift should be applied on unpropagated wavefronts: put 'q' value to 0.0 in the previous optical element")

            crl_delta = optical_element.refractive_index
            crl_w_mirr_2D_values = [OWThicknessErrorPhaseShift.h5_readsurface(thickness_error_file) for thickness_error_file in self.crl_error_profiles]

            # TO WOFRY
            generic_wavefront = input_wavefront.toGenericWavefront()

            for thickness_error_profile in crl_w_mirr_2D_values:
                phase_shift = OWThicknessErrorPhaseShift.get_crl_phase_shift(thickness_error_profile, crl_delta, generic_wavefront, self.crl_scaling_factor)

                generic_wavefront.add_phase_shift(phase_shift, Polarization.SIGMA)
                generic_wavefront.add_phase_shift(phase_shift, Polarization.PI)

            # TO SRW
            output_wavefront     = SRWWavefront.fromGenericWavefront(generic_wavefront)

            output_wavefront.Rx  = input_wavefront.Rx
            output_wavefront.Ry  = input_wavefront.Ry
            output_wavefront.dRx  = input_wavefront.dRx
            output_wavefront.dRy  = input_wavefront.dRy
            output_wavefront.xc  = input_wavefront.xc
            output_wavefront.yc = input_wavefront.yc
            output_wavefront.avgPhotEn  = input_wavefront.avgPhotEn
            output_wavefront.presCA = input_wavefront.presCA
            output_wavefront.presFT  = input_wavefront.presFT
            output_wavefront.unitElFld  = input_wavefront.unitElFld
            output_wavefront.arElecPropMatr  = copy.deepcopy(input_wavefront.arElecPropMatr)
            output_wavefront.arMomX  = copy.deepcopy(input_wavefront.arMomX)
            output_wavefront.arMomY  = copy.deepcopy(input_wavefront.arMomY)
            output_wavefront.arWfrAuxData  = copy.deepcopy(input_wavefront.arWfrAuxData)
            output_wavefront.partBeam = copy.deepcopy(input_wavefront.partBeam)

            output_wavefront.setScanningData(input_wavefront.scanned_variable_data)

            output_srw_data = SRWData(srw_beamline=srw_beamline, srw_wavefront=output_wavefront)

            self.progressBarSet(50)

            self.initializeTabs()

            tickets = []

            self.run_calculation_for_plots(output_wavefront=output_wavefront, tickets=tickets, progress_bar_value=50)

            self.plot_results(tickets, 80)

            self.progressBarFinished()
            self.setStatusMessage("")

            self.send("SRWData", output_srw_data)
            self.send("Trigger", TriggerIn(new_object=True))

        except Exception as e:
            QMessageBox.critical(self, "Error", str(e.args[0]), QMessageBox.Ok)

            self.setStatusMessage("")
            self.progressBarFinished()

            if self.IS_DEVELOP: raise e

    def run_calculation_for_plots(self, output_wavefront, tickets, progress_bar_value):
        if self.view_type==2:
            e, h, v, i = output_wavefront.get_intensity(multi_electron=False, polarization_component_to_be_extracted=PolarizationComponent.LINEAR_HORIZONTAL)

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, i[int(e.size/2)]))

            self.progressBarSet(progress_bar_value)

            e, h, v, i = output_wavefront.get_intensity(multi_electron=False, polarization_component_to_be_extracted=PolarizationComponent.LINEAR_VERTICAL)

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, i[int(e.size/2)]))

            e, h, v, p = output_wavefront.get_phase(polarization_component_to_be_extracted=PolarizationComponent.LINEAR_HORIZONTAL)

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, p[int(e.size/2)]))

            self.progressBarSet(progress_bar_value + 10)

            e, h, v, p = output_wavefront.get_phase(polarization_component_to_be_extracted=PolarizationComponent.LINEAR_VERTICAL)

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, p[int(e.size/2)]))

        elif self.view_type==1:
            e, h, v, i = output_wavefront.get_intensity(multi_electron=False)

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, i[int(e.size/2)]))

            self.progressBarSet(progress_bar_value)

            e, h, v, p = output_wavefront.get_phase()

            tickets.append(SRWPlot.get_ticket_2D(h*1000, v*1000, p[int(e.size/2)]))

        self.progressBarSet(progress_bar_value + 10)

    def propagate_new_wavefront(self, trigger):
        try:
            if trigger and trigger.new_object == True:
                if trigger.has_additional_parameter("variable_name"):
                    if self.input_srw_data is None: raise Exception("No Input Data")

                    variable_name = trigger.get_additional_parameter("variable_name").strip()
                    variable_display_name = trigger.get_additional_parameter("variable_display_name").strip()
                    variable_value = trigger.get_additional_parameter("variable_value")
                    variable_um = trigger.get_additional_parameter("variable_um")

                    if "," in variable_name:
                        variable_names = variable_name.split(",")

                        for variable_name in variable_names:
                            setattr(self, variable_name.strip(), variable_value)
                    else:
                        setattr(self, variable_name, variable_value)

                    self.input_srw_data.get_srw_wavefront().setScanningData(SRWWavefront.ScanningData(variable_name, variable_value, variable_display_name, variable_um))
                    self.propagate_wavefront()

        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

    def check_data(self):
        if len(self.crl_error_profiles) == 0: raise ValueError("No Thickness error profile specified")
        congruence.checkPositiveNumber(self.crl_scaling_factor, "Thickness Error Scaling Factor")

    @classmethod
    def h5_readsurface(cls, filename):
        x_coords, y_coords, z_values = OU.read_surface_file(filename)

        return ScaledMatrix(x_coords, y_coords, z_values.T)

    @classmethod
    def get_crl_phase_shift(cls, thickness_error_profile, crl_delta, wavefront, crl_scaling_factor=1.0):
        coord_x = thickness_error_profile.x_coord
        coord_y = thickness_error_profile.y_coord
        thickness_error = thickness_error_profile.z_values

        interpolator = RectBivariateSpline(coord_x, coord_y, thickness_error, bbox=[None, None, None, None], kx=1, ky=1, s=0)

        wavelength = wavefront.get_wavelength()
        wavefront_coord_x = wavefront.get_coordinate_x()
        wavefront_coord_y = wavefront.get_coordinate_y()

        thickness_error = interpolator(wavefront_coord_x, wavefront_coord_y)
        thickness_error[numpy.where(thickness_error==numpy.nan)] = 0.0
        thickness_error *= crl_scaling_factor

        return -2*numpy.pi*crl_delta*thickness_error/wavelength

    def getVariablesToPlot(self):
        if self.view_type == 2:
            return [[1, 2], [1, 2], [1, 2], [1, 2]]
        else:
            return [[1, 2], [1, 2]]

    def getWeightedPlots(self):
        if self.view_type == 2:
            return [False, False, True, True]
        else:
            return [False, True]

    def getWeightTickets(self):
        if self.view_type == 2:
            return [nan, nan, 0, 1]
        else:
            return [nan, 0]

    def getTitles(self, with_um=False):
        if self.view_type == 2:
            if with_um: return ["Intensity SE \u03c0 [ph/s/.1%bw/mm\u00b2]",
                                "Intensity SE \u03c3 [ph/s/.1%bw/mm\u00b2]",
                                "Phase SE \u03c0 [rad]",
                                "Phase SE \u03c0 [rad]"]
            else: return ["Intensity SE \u03c0",
                          "Intensity SE \u03c3",
                          "Phase SE \u03c0",
                          "Phase SE \u03c3"]
        else:
            if with_um: return ["Intensity SE [ph/s/.1%bw/mm\u00b2]",
                                "Phase SE [rad]"]
            else: return ["Intensity SE",
                          "Phase SE"]

    def getXTitles(self):
        if self.view_type == 2:
            return ["X [\u03bcm]", "X [\u03bcm]", "X [\u03bcm]", "X [\u03bcm]"]
        else:
            return ["X [\u03bcm]", "X [\u03bcm]"]

    def getYTitles(self):
        if self.view_type == 2:
            return ["Y [\u03bcm]", "Y [\u03bcm]", "Y [\u03bcm]", "Y [\u03bcm]"]
        else:
            return ["Y [\u03bcm]", "Y [\u03bcm]"]

    def getXUM(self):
        if self.view_type == 2:
            return ["X [\u03bcm]", "X [\u03bcm]", "X [\u03bcm]", "X [\u03bcm]"]
        else:
            return ["X [\u03bcm]", "X [\u03bcm]"]

    def getYUM(self):
        if self.view_type == 2:
            return ["Y [\u03bcm]", "Y [\u03bcm]", "Y [\u03bcm]", "Y [\u03bcm]"]
        else:
            return ["Y [\u03bcm]", "Y [\u03bcm]"]

    def callResetSettings(self):
        if ConfirmDialog.confirmed(parent=self, message="Confirm Reset of the Fields?"):
            try:
                self.resetSettings()
            except:
                pass



import os, sys

import numpy
from PyQt5.QtCore import QRect
from PyQt5.QtWidgets import QApplication, QMessageBox, QFileDialog

from matplotlib import cm
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence

import oasys.util.oasys_util as OU

try:
    from mpl_toolkits.mplot3d import Axes3D  # necessario per caricare i plot 3D
except:
    pass

class OWThicknessFileReader(OWWidget):
    name = "Thickness File Reader"
    id = "thickness_file_reader"
    description = "Thickness File Reader"
    icon = "icons/thickness_reader.png"
    author = "Luca Rebuffi"
    maintainer_email = "lrebuffi@anl.gov"
    priority = 3
    category = ""
    keywords = ["thickness_file_reader"]

    outputs = [{"name":"Thickness Error Files",
                "type":list,
                "doc":"Thickness Error Files",
                "id":"Thickness Error Files"}]

    want_main_area = 1
    want_control_area = 1

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 800
    IMAGE_HEIGHT = 610

    CONTROL_AREA_WIDTH = 405

    data=None

    separator = Setting(0)
    skip_rows = Setting(0)
    conversion_to_m_z = Setting(1.0)
    conversion_to_m_xy = Setting(1.0)

    surface_file_names = Setting(["thickness.dat"])

    negate = Setting(0)

    def __init__(self):
        super().__init__()

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               round(min(geom.width() * 0.98, self.MAX_WIDTH)),
                               round(min(geom.height() * 0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        gui.separator(self.controlArea)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Read Thickness", callback=self.read_surface)
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Render Thickness", callback=self.render_surface)
        button.setFixedHeight(45)

        input_box_l = oasysgui.widgetBox(self.controlArea, "Input", addSpace=True, orientation="vertical", height=460, width=self.CONTROL_AREA_WIDTH)

        gui.button(input_box_l, self, "Select Thickness Error Profile Data Files", callback=self.select_files)

        self.files_area = oasysgui.textArea(height=250)

        self.refresh_files_text_area()

        input_box_l.layout().addWidget(self.files_area)


        gui.comboBox(input_box_l, self, "separator", label="Separator", labelWidth=350,
                     items=["Comma", "Space"], sendSelectedValue=False, orientation="horizontal")

        oasysgui.lineEdit(input_box_l, self, "skip_rows", label="Skip Rows", labelWidth=350, orientation="horizontal", valueType=int)

        oasysgui.lineEdit(input_box_l, self, "conversion_to_m_z", label="Thickness conversion to m", labelWidth=300, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(input_box_l, self, "conversion_to_m_xy", label="Coordinates conversion to m", labelWidth=300, orientation="horizontal", valueType=float)

        gui.comboBox(input_box_l, self, "negate", label="Invert Surface", labelWidth=350,
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal")


        main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(main_tabs, "Thickness Error Surfaces")

        self.tab = []
        self.tabs = oasysgui.tabWidget(plot_tab)

        self.initialize_figures()

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)


    def initialize_figures(self):
        current_tab = self.tabs.currentIndex()

        size = len(self.tab)
        for index in range(size):
            self.tabs.removeTab(size-1-index)

        files = []

        for surface_file_name in self.surface_file_names:
            files.append(os.path.basename(surface_file_name))

        if not len(files) == 0:
            self.figures = [None]*len(files)
            self.axes    = [None]*len(files)
            self.tab     = []

            for title in files:
                self.tab.append(oasysgui.createTabPage(self.tabs, title))

            for tab in self.tab:
                tab.setFixedHeight(self.IMAGE_HEIGHT)
                tab.setFixedWidth(self.IMAGE_WIDTH)

            self.tabs.setCurrentIndex(current_tab)
        else:
            self.figures = None
            self.axes    = None

    def select_files(self):
        files, _ = QFileDialog.getOpenFileNames(self,
                                                "Select Thickness Error Profiles", "", "Data Files (*.txt *.dat)",
                                                options=QFileDialog.Options())
        if files:
            self.surface_file_names = files

            self.refresh_files_text_area()

    def refresh_files_text_area(self):
        text = ""

        for file in self.surface_file_names:
            text += file + "\n"

        self.files_area.setText(text)

    def write_thickness_files(self, error_profile_data_files, index, xx, yy, zz):
        zz = numpy.round(zz, 12)
        xx = numpy.round(xx, 12)
        yy = numpy.round(yy, 12)

        filename, _ = os.path.splitext(os.path.basename(self.surface_file_names[index]))

        thickness_profile_data_file = filename + ".h5"

        OU.write_surface_file(zz, xx, yy, thickness_profile_data_file)

        error_profile_data_files.append(thickness_profile_data_file)

    def plot_figures(self):
        error_profile_data_files = []

        for index in range(len(self.data)):
            if self.figures[index] is None:
                figure = Figure(figsize=(600, 600))
                figure.patch.set_facecolor('white')

                self.axes[index] = figure.add_subplot(111, projection='3d')
                self.figures[index] = FigureCanvasQTAgg(figure)

                self.tab[index].layout().addWidget(self.figures[index])

            xx = self.data[index][0]
            yy = self.data[index][1]
            zz = self.data[index][2]

            self.axes[index].clear()

            x_to_plot, y_to_plot = numpy.meshgrid(xx, yy)

            self.axes[index].plot_surface(x_to_plot*1e6, y_to_plot*1e6, zz*1e6,
                                          rstride=1, cstride=1, cmap=cm.coolwarm, linewidth=0.5, antialiased=True)

            self.axes[index].set_xlabel("X [\u03bcm]")
            self.axes[index].set_ylabel("Y [\u03bcm]")
            self.axes[index].set_zlabel("Z [\u03bcm]")
            self.axes[index].mouse_init()

            self.write_thickness_files(error_profile_data_files, index, xx, yy, zz)

        self.send("Thickness Error Files", error_profile_data_files)


    def send_data(self):
        error_profile_data_files = []

        for index in range(len(self.data)):
            xx = self.data[index][0]
            yy = self.data[index][1]
            zz = self.data[index][2]

            self.write_thickness_files(error_profile_data_files, index, xx, yy, zz)

        self.send("Thickness Error Files", error_profile_data_files)

    def read_surface(self):
        try:
            self.read_data_files()
            self.send_data()

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 exception.args[0],
                                 QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception



    def render_surface(self):
        try:
            self.read_data_files()
            self.plot_figures()

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 exception.args[0],
                                 QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception


    def read_data_files(self):
        self.data = []

        for surface_file_name in self.surface_file_names:
            surface_file_name = congruence.checkDir(surface_file_name)

            data = numpy.loadtxt(surface_file_name, delimiter="," if self.separator==0 else " ", skiprows=self.skip_rows)

            xx = numpy.unique(data[:, 0]) * self.conversion_to_m_xy
            yy = numpy.unique(data[:, 1]) * self.conversion_to_m_xy
            zz = numpy.reshape(data[:, 2], (len(xx), len(yy))).T

            zz = zz * self.conversion_to_m_z if self.negate == 0 else -1.0 * zz * self.conversion_to_m_z

            self.data.append([xx, yy, zz])

        self.initialize_figures()
